//
//  CalculatorBrain.swift
//  CalculatorBrain
//
//  Created by Carlos Barrera on 23/10/17.
//  Copyright © 2017 Carlos Barrera. All rights reserved.
//
// Created following Stanford Course (https://itunes.apple.com/us/course/developing-ios-10-apps-with-swift/id1198467120)


import Foundation


public class CalculatorBrain {
    
    private var accumulator = 0.0
    private var history: [String] = []
    private var lastOperation :LastOperation = .Clear
    
    private var operations: Dictionary<String, Operation> = [
        "π": Operation.Constant(Double.pi),
        "e": Operation.Constant(M_E),
        "√": Operation.UnaryOperation(sqrt),
        "*": Operation.BinaryOperation({ $0 * $1 }),
        "-": Operation.BinaryOperation({ $0 - $1 }),
        "/": Operation.BinaryOperation({ $0 / $1 }),
        "+": Operation.BinaryOperation({ $0 + $1 }),
        "=": Operation.Equals,
        "c": Operation.Clear
    ]
    
    private enum Operation {
        case Constant(Double)
        case UnaryOperation((Double) -> Double)
        case BinaryOperation((Double, Double) -> Double)
        case Equals
        case Clear
    }
    
    private enum LastOperation {
        case Digit
        case Constant
        case UnaryOperation
        case BinaryOperation
        case Equals
        case Clear
    }
    public func insertConstants(_ constants: [String:Any]){
        
        var operations = self.operations
        for constant in constants {
            if let value = Double(String("\(constant.value)".split(separator: " ")[0])){
                // we need to know the type of the obj
                // For the moment we assume we'll pass a measurements, and so the left side of the string is the value
                operations.updateValue(.Constant(value), forKey: constant.key)
            }
        }
        self.operations = operations
        
    }
    
    public func insertConstants(_ constants: [String:Double]){
        var operations = self.operations
        for constant in constants {
            operations.updateValue(.Constant(constant.value), forKey: constant.key)
        }
        self.operations = operations
    }

    func setOperand(operand: Double) {
        if lastOperation == .UnaryOperation {
            history.removeAll()
        }
        
        accumulator = operand
        history.append(String(operand))
        lastOperation = .Digit
    }
    
    func performOperand(symbol: String) {
        if let operation = operations[symbol] {
//            print("Identified as operation: \(symbol)")
            switch operation {
            case .Constant(let value):
                history.append(symbol)
                accumulator = value
                lastOperation = .Constant
            case .UnaryOperation(let function):
                wrapWithParens(symbol: symbol)
                accumulator = function(accumulator)
                lastOperation = .UnaryOperation
            case .BinaryOperation(let function):
                if lastOperation == .Equals {
                    history.removeLast()
                }
                history.append(symbol)
                executePendingBinaryOperation()
                pending = PendingBinaryOperationInfo(binaryFunction: function, firstOperand: accumulator)
                lastOperation = .BinaryOperation
            case .Equals:
                if lastOperation == .BinaryOperation {
                    history.append(String(accumulator))
                }
                history.append(symbol)
                executePendingBinaryOperation()
                lastOperation = .Equals
            case .Clear:
                clear()
                lastOperation = .Clear
            }
        } else {
//            print("Identified as value: \(symbol)")
            setOperand(operand: Double(symbol)!)
        }
        
    }
    
    var result: Double {
        get {
            return accumulator;
        }
    }
    
    var isPartialResult: Bool {
        get {
            return pending != nil
        }
    }
    
    var description: String {
        get {
            return history.joined(separator: " ")
        }
    }
    
    private func wrapWithParens(symbol: String) {
        if lastOperation == .Equals {
            history.insert(")", at: history.count - 1)
            history.insert(symbol, at: 0)
            history.insert("(", at: 1)
        } else {
            history.insert(symbol, at: history.count - 1)
            history.insert("(", at: history.count - 1)
            history.insert(")", at: history.count)
        }
    }
    
    private func executePendingBinaryOperation() {
        if pending != nil {
            accumulator = pending!.binaryFunction(pending!.firstOperand, accumulator)
            pending = nil
        }
    }
    
    private func clear() {
        accumulator = 0
        pending = nil
        history.removeAll()
        lastOperation = .Clear
    }
    
    private var pending: PendingBinaryOperationInfo?
    
    private struct PendingBinaryOperationInfo {
        var binaryFunction: (Double, Double) -> Double
        var firstOperand: Double
    }
}
