//
//  Serie.swift
//  MeasurementHelper
//
//  Created by Carlos Barrera on 3/11/17.
//  Copyright © 2017 Carlos Barrera. All rights reserved.
//

import Foundation

class Serie : Codable {
    
    var rounds = [Round]()
    
    private var sharedInputs = [Input]()
    
    var results = [Input]()
    
    var regression = Regression()
    
    var reactants = [String:Solution]()
    
    
    public func addRound(){
        self.rounds.append(Round())
    }
    
    
    public func addInput(_ input: Input, toRound: Int = -1) {
        let sharedIn = input.shared
        switch sharedIn {
        case false:
            switch toRound {
            case -1:
                // add the new input to the last round
                self.addRound()
                var round = self.rounds.popLast()!
                round.addInput(input)
                self.rounds.append(round)
                
                
            default:
                if self.rounds.count == 0 {
                    self.addRound()
                }
                // add the input to the desired round
                if toRound < self.rounds.count - 1{
                    var round = self.rounds[toRound]
                    round.addInput(input)
                }
            }
            
        default:
            self.sharedInputs.append(input)
        }
    }
    
    public func getInputs()-> [Input]{
        var inputs = [Input]()
        for shared in self.sharedInputs {
            inputs.append(shared)
        }
        for round in self.rounds {
            for (_, value) in round.inputs {
                for input in value {
                    inputs.append(input)
                }
                
            }
        }
        return inputs
    }
    
    public func cleanShared(){
        self.sharedInputs = [Input]()
    }
    
    public func addResult(_ value: Double, forVariable variable: Variable ) {
        let input = Input.init(fromVariable: variable, withValue: value)
        self.results.append(input)
    }
    
    public func cleanResults() {
        self.results = [Input]()
    }
    
    private func getResultsWithExperiment(experiment: Experiment){
        let formula = experiment.formula
        // empty the results
        self.cleanResults()
        let shared = self.sharedInputs
        let notSharedVariablesLabels = experiment.variables.filter({$0.shared == false}).map({$0.label})
        
        // we must create a result for each round
        for round in self.rounds {
            var inputsToCalculate = [Input]()
            inputsToCalculate.append(contentsOf: shared)
            
            // add all inputs for all non shared variables
            for label in notSharedVariablesLabels {
                inputsToCalculate.append(contentsOf:round.getInputsForLabel(label))
            }
            
            let brain = CalculatorBrain()
            
            let result = self.getResultFor(formula: formula, inputs: inputsToCalculate, brain: brain)
            self.addResult(result, forVariable: experiment.result!)
        }
        
    }
    
    public func getResultValuesForExperiment(experiment: Experiment) -> [Double]{
        self.getResultsWithExperiment(experiment: experiment)
        return self.results.map({$0.measure.value})
    }
    public func getResultValuesForRegression(formula: [Step], variables: [String:Double], withInputs values: [Double]) -> [Double]{
        // empty the results
        
        
        var results = [Double]()
        // we must create a result for each round
        for value in values {
            let brain = CalculatorBrain()
            brain.insertConstants(variables)
            brain.insertConstants(["Y":value])
            for step in formula {
                brain.performOperand(symbol: step.operation)
            }

            let result = brain.result
            results.append(result)
            
        }
        
        return results
    }

    
    public func getShared() -> [Input]{
        return self.sharedInputs
    }
    
    public func getRounds() -> [Round]{
        return self.rounds
    }
    
    private func insertVariables(inputs: [Input], brain:CalculatorBrain) {
        for input in inputs {
            brain.insertConstants([input.label:input.measure])
        }
    }
    
    
    private func getResultFor(formula:[Step], inputs:[Input], brain: CalculatorBrain) -> Double {
        self.insertVariables(inputs: inputs, brain:brain)
        
        for step in formula {
            brain.performOperand(symbol: step.operation)
        }
        
        return brain.result
        
    }
    
    //MARK: - Stats accessors
    public func getMean() -> Double?{
        // the Sigma average function takes into account the possibility of nil result
        
        var values = [Double]()
        for input in self.results {
            values.append(input.measure.value)
        }
        return Sigma.average(values)
    }
    
    public func getStdDev() -> Double?{
        // the Sigma average function takes into account the possibility of nil result
        var values = [Double]()
        for input in self.results {
            values.append(input.measure.value)
        }
        return Sigma.standardDeviationSample(values)
        
    }
    public func getSEM() -> Double?{
        // the Sigma average function takes into account the possibility of nil result
        var values = [Double]()
        for input in self.results {
            values.append(input.measure.value)
        }
        return Sigma.standardErrorOfTheMean(values)
        
    }

    
    

    
}
