//
//  MeasurementHelperTests.swift
//  MeasurementHelperTests
//
//  Created by Carlos Barrera on 22/10/17.
//  Copyright © 2017 Carlos Barrera. All rights reserved.
//

import XCTest
@testable import MeasurementHelper

class MH_InitializationTests: XCTestCase {
    
    
    override func setUp() {
        super.setUp()
        // Put setup code here. This method is called before the invocation of each test method in the class.
    }
    
    override func tearDown() {
        // Put teardown code here. This method is called after the invocation of each test method in the class.
        super.tearDown()
    }
    
//    func testExample() {
//        // This is an example of a functional test case.
//        // Use XCTAssert and related functions to verify your tests produce the correct results.
//        
//        
//        
//    }
    
    //MARK: - Constants
    static let points = [Datapoint(1.0, 3.0),Datapoint(2.4,6.0), Datapoint(5.4,15.0), Datapoint(7.2,24.0), Datapoint(9.0, 28.7)]
    static let slopeReferenceValue = 3.3364
    static let interceptReferenceValue = -1.3419
    static let rCoeffReferenceValue = 0.9878

    
    
    // MARK: - Session tests
    func testEmptySessionCreation() {
        let session = Session()
        XCTAssert(session.currentSerie == -1, "Session created with incorrect currentSerie -> \(session.currentSerie)")
    }
    
    func testCompareSessions() {
        let session1 = Session()
        let session2 = Session()
        session2.startDate = session1.startDate
        XCTAssertEqual(session1, session2, "Unable to detect equal sessions")
    }
    
    func testEncodeSessionToString() {
        let session = Session()
        
        let string = session.exportString()
        let startDate = session.startDate
        let formatter = ISO8601DateFormatter()
        let dateToCompare = formatter.string(from: startDate)

        let expectedString = """
        {"experimentType":{"formula":[],"regressionType":false,"name":"","description":"","variables":[]},"series":[],"currentSerie":-1,"startDate":"\(dateToCompare)","description":"Description"}
        """
        
        XCTAssertEqual(string, expectedString, "String encoding incorrect, obtained \(string) instead of \(expectedString)")
    }
    
    func testDecodeSessionFromString() {
        let session = Session()
        let startDate = session.startDate
        let formatter = ISO8601DateFormatter()
        let dateToCompare = formatter.string(from: startDate)
        let nameString = "Very Cool Experiment"
        let expTypeDescriptionString = "Let's analyze something nice for a change"
        let sessionDescriptionString = "Have fun"
        
        let encodedString = """
        {"experimentType":{"formula":[],"regressionType":false,"name":"\(nameString)","description":"\(expTypeDescriptionString)","variables":[]},"series":[],"currentSerie":-1,"startDate":"\(dateToCompare)","description":"\(sessionDescriptionString)"}
        """
        session.decodeFrom(string: encodedString)
        
        XCTAssertEqual(formatter.string(from: session.startDate), formatter.string(from: startDate), "Incorrectly decoded startDate, obtained \(session.startDate) instead of \(startDate)")
        
        XCTAssertEqual(session.series.count, 0, "Incorrectly decoded series, obtained \(session.series.count) instead of 0")
        XCTAssertEqual(session.description, sessionDescriptionString, "Incorrectly decoded session description, obtained \(session.description) instead of \(sessionDescriptionString)")
    }
    
    func testSessionCreationWithExpType() {
        let sessionName = "Wow what a wonderfull experiment"
        let experimentType = Experiment()
        let session = Session.newSessionWithExperimentType(experimentType, name: sessionName)
        
        XCTAssertEqual(session.description, sessionName, "Incorrectly transferring the name when creating the session, obtained \(session.description) instead of \(sessionName)")
    }
    
    func testSeriesManagement() {
        let session = Session()
        let count = 4
        for _ in 0..<count {
            session.addSeries()
        }
        XCTAssertEqual(session.series.count , count, "Incorrectly adding series, obtained \(session.series.count) instead of \(count)")
    }
    
    func testSeriesSwitching() {
        let session = Session()
        let count = 4
        for _ in 0..<count {
            session.addSeries()
        }
        let destination = count - 1
        
        XCTAssertTrue( session.switchToSeries(newSeries: destination))
        
        XCTAssertEqual(session.currentSerie, destination, "Not swithcing series correctly, went to \(session.currentSerie) instead of \(destination)")
    }
    
    func testAddInput(){
        let session = Session()
        let input = Input()
        session.addInput(input)
        let inputCount = session.series.first!.getInputs().count
        let expectedCount = 1
        XCTAssertEqual(inputCount, expectedCount, "Incorrect count of inputs when adding, obtained \(inputCount) instead of \(expectedCount)")
        
    }
    
    func testSharedManagement() {
        let session = Session()
        let input = Input()
        // add one not shared input
        session.addInput(input)
        // set the input as shared
        input.shared = true
        // add the input as shared
        session.addInput(input)
        // Clean shared inputs
        session.cleanShared()
        let inputCount = session.series.first!.getInputs().count
        let expectedCount = 1
        XCTAssertEqual(inputCount, expectedCount, "Incorrect count of inputs when cleaning shared, obtained \(inputCount) instead of \(expectedCount)")

        
        
    }

    
   
    
    // MARK: - Experiment tests
    func testEmptyExperimentCreation() {
        let experiment = Experiment()
        XCTAssert(experiment.name == "", "Experiment created with incorrect name -> \(experiment.name)")
        XCTAssert(experiment.description == "", "Experiment created with incorrect description -> \(experiment.description)")
        XCTAssert(experiment.formula.count == 0, "Experiment created with incorrect formula -> \(experiment.formula)")
        XCTAssert(experiment.variables.count == 0, "Experiment created with incorrect variables -> \(experiment.variables)")
        XCTAssertNil(experiment.valorant, "Experiment created with incorrect valorant -> \(experiment.valorant.debugDescription)")
        XCTAssertNil(experiment.result, "Experiment created with incorrect result variable -> \(experiment.result.debugDescription)")
    }
    
    func testCompareExperiments() {
        let exp1 = Experiment()
        let exp2 = Experiment()
        
        XCTAssertEqual(exp1, exp2, "When they are equal we report them as different")
        exp2.description = "What a difference it makes"
        
        XCTAssertNotEqual(exp1, exp2, "When they are different, we report them as equal")
    }
    
    func testFormulaParsingFromSteps() {
        let experiment = Experiment()
        let formulaExpected = "2 * X + 4.5 ="
        experiment.formula = Experiment.stepsFromFormula(formula: formulaExpected)
        
        let calculatedFormulaString = experiment.formulaString
        
        XCTAssertEqual(formulaExpected, calculatedFormulaString, "Formula conversion from steps not correct \(calculatedFormulaString) instead of \(formulaExpected)")
    }

    
    // MARK: - Step tests
    func testEmptyStepCreation() {
        let order = 0
        let operation = ""
        
        let step = Step()
        
        XCTAssertEqual(step.order, order, "Incorrect order transfer -> \(step.order)")
        XCTAssertEqual(step.operation, operation, "Incorrect operation transfer -> \(step.operation)")
        XCTAssertEqual(step.description(), "\(order):\(operation)", "Incorrect description generation -> \(step.description())")
    }
    
    func testFilledStepCreation() {
        let order = 3
        let operation = "product"
        
        let step = Step(operation: operation, atIndex: order)
        
        XCTAssertEqual(step.order, order, "Incorrect order transfer -> \(step.order)")
        XCTAssertEqual(step.operation, operation, "Incorrect operation transfer -> \(step.operation)")
        XCTAssertEqual(step.description(), "\(order):\(operation)", "Incorrect description generation -> \(step.description())")
    }
    

    
    // MARK: - Variable tests
    func testMinimalVariableCreation() {
        let labelString = "label"
        let dimension = SupportedDimensions.mass
        let shared = false
        let kind = Kind.other
        
        let variable = Variable(label: labelString, dim: dimension, shared: shared, kind: kind)
        XCTAssert(variable.label == labelString, "Variable label incorrectly transferred -> \(variable.label)")
        XCTAssert(variable.dimension == dimension, "Variable dimension incorrectly transferred -> \(variable.dimension)")
        XCTAssert(variable.shared == shared, "Variable shared incorrectly transferred -> \(variable.shared)")
        XCTAssert(variable.kind == kind, "Variable kind incorrectly transferred -> \(variable.kind)")
        XCTAssert(variable.descriptive == labelString, "Variable descriptive not equal to label-> \(variable.descriptive)")
        XCTAssertFalse(variable.paired, "Variable paired not false-> \(variable.paired)")
    }
    
    func testEmptyVariableCreation() {
        let variable = Variable()
        XCTAssert(variable.label == "", "Variable label incorrectly transferred -> \(variable.label)")
        XCTAssert(variable.dimension == .mass, "Variable dimension incorrectly transferred -> \(variable.dimension)")
        XCTAssert(variable.shared == true, "Variable shared incorrectly transferred -> \(variable.shared)")
        XCTAssert(variable.kind == .acid, "Variable kind incorrectly transferred -> \(variable.kind)")
        XCTAssert(variable.descriptive == "", "Variable descriptive not equal to label-> \(variable.descriptive)")
        XCTAssertFalse(variable.paired, "Variable paired not false-> \(variable.paired)")

    }
    

    
    // MARK: - Serie tests
    func testEmptySeriesCreation() {
        let rounds = [Round]()
//        let sharedInputs = [Input]()
        let results = [Input]()
        let regression = Regression()
        let reactants = [String:Solution]()
        
        let serie = Serie()
        
        XCTAssertEqual(serie.rounds.count, rounds.count, "Incorrect rounds initialization -> \(serie.rounds)")
        XCTAssertEqual(serie.results.count, results.count, "Incorrect results initialization -> \(serie.results)")
        XCTAssertEqual(serie.regression, regression, "Incorrect regression initialization -> \(serie.regression)")
        XCTAssertEqual(serie.reactants.count, reactants.count, "Incorrect reactants initialization -> \(serie.reactants)")

    }
    

    
    // MARK: - Input tests
    func testCreateMassInput() {
        let value = 2.0
        let unit = UnitMass.grams
        let label = "Weight"
        
        let measurement = Measurement(value: value, unit: unit)
        let input = Input(label: label, measure: measurement)
        
        XCTAssert(input.label == label, "Label value incorrectly transferred -> \(input.label)")
        XCTAssert(input.shared == false, "Shared value incorrectly initialized -> \(input.shared)")
        XCTAssert(input.measure.value == value, "MeasurementValue incorrectly transferred -> \(input.measure.value)")
        XCTAssert(input.measure.unit == unit, "MeasurementUnit incorrectly transferred -> \(input.measure.unit.description)")
        XCTAssertEqual(input.measure.description, measurement.description, "Measurement constructed not correct -> \(input.measure.description)")

    }

    func testCreateVolumeInput() {
        let value = 2.0
        let unit = UnitVolume.milliliters
        let label = "Volume"
        
        let measurement = Measurement(value: value, unit: unit)
        let input = Input(label: label, measure: measurement)
        
        XCTAssert(input.label == label, "Label value incorrectly transferred -> \(input.label)")
        XCTAssert(input.shared == false, "Shared value incorrectly initialized -> \(input.shared)")
        XCTAssert(input.measure.value == value, "MeasurementValue incorrectly transferred -> \(input.measure.value)")
        XCTAssert(input.measure.unit == unit, "MeasurementUnit incorrectly transferred -> \(input.measure.unit.description)")
        XCTAssertEqual(input.measure.description, measurement.description, "Measurement constructed not correct -> \(input.measure.description)")
        
    }
    func testCreateConcentrationInput() {
        let value = 2.0
        let unit = UnitConcentrationMass.molar
        let label = "Molarity"
        
        let measurement = Measurement(value: value, unit: unit)
        let input = Input(label: label, measure: measurement)
        
        XCTAssert(input.label == label, "Label value incorrectly transferred -> \(input.label)")
        XCTAssert(input.shared == false, "Shared value incorrectly initialized -> \(input.shared)")
        XCTAssert(input.measure.value == value, "MeasurementValue incorrectly transferred -> \(input.measure.value)")
        XCTAssert(input.measure.unit == unit, "MeasurementUnit incorrectly transferred -> \(input.measure.unit.description)")
        XCTAssertEqual(input.measure.description, measurement.description, "Measurement constructed not correct -> \(input.measure.description)")
        
    }

    func testCreateDispersionInput() {
        let value = 2.0
        let unit = UnitDispersion.percent
        let label = "%"
        
        let measurement = Measurement(value: value, unit: unit)
        let input = Input(label: label, measure: measurement)
        
        XCTAssert(input.label == label, "Label value incorrectly transferred -> \(input.label)")
        XCTAssert(input.shared == false, "Shared value incorrectly initialized -> \(input.shared)")
        XCTAssert(input.measure.value == value, "MeasurementValue incorrectly transferred -> \(input.measure.value)")
        XCTAssert(input.measure.unit == unit, "MeasurementUnit incorrectly transferred -> \(input.measure.unit.description)")
        XCTAssertEqual(input.measure.description, measurement.description, "Measurement constructed not correct -> \(input.measure.description)")
        
    }

    
    func testCreateFromVariable() {
        let label = "Weight"
        let dimension = SupportedDimensions.mass
        let value = 2.0
        let shared = false
        let unit = UnitMass.grams
        
        let measurement = Measurement(value: value, unit: unit)

        
        let variable = Variable(label: label, dim: dimension, shared: shared)
        
        let input = Input.init(fromVariable: variable, withValue: value)
        
        XCTAssert(input.label == label, "Label value incorrectly transferred -> \(input.label)")
        XCTAssert(input.shared == false, "Shared value incorrectly initialized -> \(input.shared)")
        XCTAssert(input.measure.value == value, "MeasurementValue incorrectly transferred -> \(input.measure.value)")
        XCTAssert(input.measure.unit == unit, "MeasurementUnit incorrectly transferred -> \(input.measure.unit.description)")
        XCTAssertEqual(input.measure.description, measurement.description, "Measurement constructed not correct -> \(input.measure.description)")

    }

    
    
    // MARK: - Regression tests
    func testEmptyRegressionCreation() {
        let points = [Datapoint]()
        let dataUnits = [String:String]()
        let dataLabels = [String:String]()
        let correlation = Correlation(type: .linear)
        
        let regression = Regression()
        
        XCTAssertEqual(regression.points, points, "Points are not correctly initialized ->\(regression.points)")
        XCTAssertEqual(regression.dataUnits, dataUnits, "DataUnits are not correctly initialized ->\(regression.dataUnits)")
        XCTAssertEqual(regression.dataLabels, dataLabels, "Data Labels  are not correctly initialized ->\(regression.dataLabels)")
        XCTAssertEqual(regression.correlation, correlation, "Correlation is not correctly initialized ->\(regression.correlation)")
        
        
        
    }
    
    func testRegressionModelCreationWithPoints() {

        let regression = Regression()
        regression.generateModel(data: MH_InitializationTests.points)
        let formatter = NumberFormatter()
        formatter.minimumIntegerDigits = 1
        formatter.maximumFractionDigits = 4
        let mPattern = (formatter.string(from: NSNumber(value: MH_InitializationTests.slopeReferenceValue )))!
        let cPattern = (formatter.string(from: NSNumber(value: MH_InitializationTests.interceptReferenceValue)))!
        let rCoeffPattern = (formatter.string(from: NSNumber(value: MH_InitializationTests.rCoeffReferenceValue)))!
        
        let calculatedM = (regression.correlation.components[ComponentLabels.slope])!
        let calculatedC = (regression.correlation.components[ComponentLabels.intercept])!
        let calculatedRCoef = regression.calculateLinearRegressionRcoef(slope: calculatedM, intercept: calculatedC)
        
        let calculatedMPattern = formatter.string(from: NSNumber(value: calculatedM))!
        let calculatedCPattern = formatter.string(from: NSNumber(value: calculatedC))!
        let calculatedRCoeffPattern = formatter.string(from: NSNumber(value: calculatedRCoef))!
        
        XCTAssertEqual(mPattern, calculatedMPattern, "Slope incorrect calculation-> \(calculatedMPattern) intead of \(mPattern)")
        XCTAssertEqual(cPattern, calculatedCPattern, "Slope incorrect calculation-> \(calculatedCPattern) intead of \(cPattern)")
        XCTAssertEqual(rCoeffPattern, calculatedRCoeffPattern, "R coefficient incorrect calculation-> \(calculatedRCoeffPattern) intead of \(rCoeffPattern)")

        
    }

    
    // MARK: - Correlation tests
    func testEmptyCorrelationCreation() {
        let type = Fitting.linear
        let correlation = Correlation(type: type)
        let slopeNumber = (correlation.components[ComponentLabels.slope])!
        let interceptNumber = (correlation.components[ComponentLabels.intercept])!

        let rCoeff = 0.0
        let labels = [ComponentLabels.slope, ComponentLabels.intercept]
//        let components = [[ComponentLabels.slope:0.0], [ComponentLabels.intercept:0]]
        let formula = "Y - \(ComponentLabels.intercept) / \(ComponentLabels.slope) ="
        let coefFormatter = NumberFormatter()
        coefFormatter.minimumIntegerDigits = 1
        coefFormatter.maximumFractionDigits = 3
        let description = "y = \((coefFormatter.string(from: NSNumber(value:slopeNumber)))!) * x + \((coefFormatter.string(from: NSNumber(value:interceptNumber)))!)"
        
        XCTAssertEqual(correlation.rCoefficient, rCoeff, "Incorrect rCoefficient initialization->\(correlation.rCoefficient) instead of \(rCoeff)")
        XCTAssertEqual(correlation.labels, labels, "Incorrect labels initialization->\(correlation.labels) instead of \(labels)")
        XCTAssertEqual(correlation.formula, formula, "Incorrect formula initialization->\(correlation.formula) instead of \(formula)")
        XCTAssertEqual(correlation.description, description, "Incorrect description initialization->\(correlation.description) instead of \(description)")

    }
    
    func testCorrelationConvertToFormula() {
        let type = Fitting.linear
        let correlation = Correlation(type: type)
        let slopeNumber = 2.3
        correlation.components[ComponentLabels.slope] = slopeNumber
        let interceptNumber = 3.5
        correlation.components[ComponentLabels.intercept] = interceptNumber

        let expectedSteps = Experiment.stepsFromFormula(formula: correlation.formula)
        let expectedComponents = [ComponentLabels.slope:slopeNumber, ComponentLabels.intercept:interceptNumber]
        let result = correlation.convertToFormula()
        
        let calculatedSteps = result.0
        let calculatedComponents = result.1
        
        XCTAssertEqual(calculatedSteps.map({$0.operation}).joined(separator: " "), expectedSteps.map({$0.operation}).joined(separator: " "), "Incorrect steps conversion obtained \(calculatedSteps.description) instead of \(expectedSteps.description)")
        // as dictionaries are not sorted we can't compare directly the arrays
        // let's process them a little and compare the resulting string
        let calculatedPattern = calculatedComponents.map({"\($0.key):\($0.value)"}).sorted().joined(separator: " ")
        let expectedPattern = expectedComponents.map({"\($0.key):\($0.value)"}).sorted().joined(separator: " ")

        XCTAssertEqual(calculatedPattern, expectedPattern, "Incorrect components conversion obtained \(calculatedPattern) instead of \(expectedPattern)")
    }

    
    // MARK: - DataPoint tests
    func testEmptyDatapointCreation() {
        let point = Datapoint()
        
        XCTAssert((point.x == 0.0) && (point.y == 0.0), "Empty Datapoint is not 0,0 -> (\(point.x),\(point.y))")
    }
    
    
    // MARK: - Solution tests
    func testGuessMolecularWeightInit() {
        let substanceName = "HCl"
        let newReact = Solution(substance: substanceName, atMolarity: 0.001, factor: 1.002, equivalence: 1, MW: nil)
        let rightMW = Solution.getMW(forCompound: substanceName)
        XCTAssert(newReact.molecularWeight == rightMW, "MW obtained incorrect: \(newReact.molecularWeight)")
        
    }
    
    func testForceMolecularWeightInit() {
        let substanceName = "H2SO4"
        // the MW of H2SO4 is 98.072
        let wrongMolecularWeight = 34.53
        let newSolution = Solution(substance: substanceName, atMolarity: 0.01, MW: wrongMolecularWeight)
        let rightMolecularWeight = Solution.getMW(forCompound: substanceName)
        
        XCTAssertNotEqual(rightMolecularWeight, newSolution.molecularWeight)
    }
    
    func testInvalidParametersInit() {
        let newSolution = Solution(substance: "", atMolarity: 0.0, factor: 0, equivalence: 0, MW: 0)
        
        XCTAssertEqual(newSolution.substance, "")
    }
    
    func testNoMWForInvalidSubstance() {
        let newSolution = Solution(substance: "", atMolarity: 0, MW: nil)
        
        XCTAssertEqual(newSolution.molecularWeight , 0.0)

    }
    
    func testRecoverLabelConc() {
        let substanceName = "HCl"
        let desiredMolarityLabel = 0.2

        let newSolution = Solution(substance: substanceName, atMolarity: desiredMolarityLabel, MW: nil)
        XCTAssertEqual(newSolution.labelConcentration, desiredMolarityLabel)
    }
    
    func testGetRightRealConcentration() {
        let substanceName = "NaOH"
        let initialMolarity = 0.1
        let factor = 1.002
        
        let expectedRealConcentration = 0.1002
        
        let newSolution = Solution(substance: substanceName, atMolarity: initialMolarity, factor: factor, MW: nil)
        
        let formatter = NumberFormatter()
        formatter.minimumIntegerDigits = 1
        formatter.maximumFractionDigits = 4
        
        let calculatedRealConcentrationPattern = (formatter.string(from: NSNumber(value:(newSolution.realConcentration?.value)!)))!
        let expectedRealConcentrationPattern = (formatter.string(from: NSNumber(value:expectedRealConcentration)))!
        
        
        XCTAssertEqual(calculatedRealConcentrationPattern, expectedRealConcentrationPattern, "Incorrect Real concentration calculation ->\(calculatedRealConcentrationPattern) instead of \(expectedRealConcentrationPattern)")
        
        

    }

    // MARK: - Round tests
    func testEmptyRoundCreation(){
        let round = Round()
        let desiredCount = 0
        let fakeLabel = "What?"
        let emptyInputs = round.getInputsForLabel(fakeLabel)
        let expectedInputCount = 0
        XCTAssertEqual(emptyInputs.count, expectedInputCount, "Incorrect inputs for label \"\(fakeLabel)\" ->\(emptyInputs.count) instead of \(expectedInputCount)")
        
        
        XCTAssertEqual(round.inputs.count, desiredCount, "Incorrect inputs initialization->\(round.inputs.count) instead of \(desiredCount)" )
    }
    
    func testReplicateCount() {
        let input = Input()
        let label = "IntenseValues"
        input.label = label
        var round = Round()
        let expectedReplicates = 12
        
        for value in 0..<expectedReplicates {
            input.measure.value = Double(value)
            round.addInput(input)
        }
        let realReplicates = round.replicateCount()
        
        XCTAssertEqual(realReplicates, expectedReplicates, "Incorrect number of replicates ->\(realReplicates) instead of \(expectedReplicates)")
    }
    
}
