//
//  MeasurementHelperInteractionsTests.swift
//  MeasurementHelperTests
//
//  Created by Carlos Barrera on 6/12/17.
//  Copyright © 2017 Carlos Barrera. All rights reserved.
//

import XCTest
@testable import MeasurementHelper

class MH_InteractionsTests: XCTestCase {
    
    enum Constants {
        static let valorantVolume = "titrantVolume"
        static let valorantConcentration = "titrantConcentration"
        static let sampleAcidMolecularWeight = "sampleAcidMolecularWeight"
        static let sampleWeight = "sampleWeight"
    }
    
    var session = Session()
    var solutions = [Solution]()
    
    override func setUp() {
        super.setUp()
        // Put setup code here. This method is called before the invocation of each test method in the class.
        // we'll create a complex setup of entities
        // Solutions
        // Add some solutions
        let solution1 = Solution(substance: "NaOH", atMolarity: 0.1, factor: 1.002, equivalence: 1, MW:nil, kind: .base)
        let solution2 = Solution(substance: "HCl", atMolarity: 0.01, factor: 0.998, equivalence: 1, MW:nil, kind: .acid)
        let solution3 = Solution(substance: "H2SO4", atMolarity: 0.01, factor: 1.001, equivalence: 2, MW:nil, kind: .acid)
        let solution4 = Solution(substance: "HNO3", atMolarity: 0.01, factor: 0.989, equivalence: 1, MW:nil, kind: .acid)
        let solution5 = Solution(substance: "KOH", atMolarity: 0.01, factor: 0.989, equivalence: 1, MW:nil, kind: .base)
        // create solutions collection
        solutions = [solution1, solution2, solution3, solution4, solution5]

        // Session
        // set startDate to now
        session.startDate = Date.init(timeIntervalSinceNow: 0.0)

        // Create experiment type
        let exp1 = Experiment()
        exp1.name = "Iogurt acidity"
        exp1.description = "Acid-Base titration to determine acidity of iogurt with NaOH"
        exp1.formula = Experiment.stepsFromFormula(formula: "\(Constants.valorantVolume) * \(Constants.valorantConcentration) * \(Constants.sampleAcidMolecularWeight) / \(Constants.sampleWeight) / 10 =")
        let var1 = Variable(label: Constants.valorantVolume, dim: .volume, shared: false, kind: .base)
        let var2 = Variable(label: Constants.valorantConcentration, dim: .concentration, shared: true, kind: .other)
        let var3 = Variable(label:Constants.sampleAcidMolecularWeight, dim: .concentration, shared: true, kind: .other)
        let var4 = Variable(label: Constants.sampleWeight, dim: .mass, shared: true, kind: .other)
        exp1.variables = [var1, var2, var3, var4]
        
        let result = Variable(label: "%AceticAcid", dim: .percentage, shared: false, kind: .other, description: "% of Acetic Acid in the sample", paired: false)
        exp1.result = result
        // set it as current experiment type
        session.experimentType = exp1
        let serie = Serie()
        
        serie.reactants.updateValue(solution1, forKey: solution1.substance)

        session.series.append(serie)
        
    }
    
    override func tearDown() {
        // Put teardown code here. This method is called after the invocation of each test method in the class.
        super.tearDown()
    }
    
    func testOneSerieOneRoundForIogurt() {
        // This is an example of a functional test case.
        // Use XCTAssert and related functions to verify your tests produce the correct results.
        let weightInGrams = 8.5
        let concentrationValue = solutions[0].realConcentration!
        let molecularWeight = 90.0
        let volumeUsed = 2.4
        let expectedResult = [0.254626]
        
        let serie = session.series.first!

        let variables = session.experimentType.variables
        for variable in variables {
            var input = Input()
            switch variable.label{
            case Constants.sampleAcidMolecularWeight:
                input = Input(fromVariable: variable, withValue: molecularWeight)
            case Constants.sampleWeight:
                input = Input(fromVariable: variable, withValue: weightInGrams)
            case Constants.valorantConcentration:
                input = Input(fromVariable: variable, withValue: concentrationValue.value)
            case Constants.valorantVolume:
                input = Input(fromVariable: variable, withValue: volumeUsed)
            
            default:
                break
            }
            serie.addInput(input)

        }
        // here we have introduced the values for the experiment and can request results
        let results = serie.getResultValuesForExperiment(experiment: session.experimentType)
        // Customary formatter for results
        let formatter = NumberFormatter()
        formatter.minimumIntegerDigits = 1
        formatter.maximumFractionDigits = 4
        
        let resultPattern = (formatter.string(from: NSNumber(value:results[0])))!
        let expectedResultPattern = (formatter.string(from: NSNumber(value:expectedResult[0])))!
        
        
        XCTAssertEqual(resultPattern, expectedResultPattern, "Incorrect results obtained->\(resultPattern) instead of \(expectedResultPattern)")
        
        
        
        
    }
    
    func testOneSerieOneRoundMultipleInputsForIogurt() {
        // This is an example of a functional test case.
        // Use XCTAssert and related functions to verify your tests produce the correct results.
        let weightInGrams = 8.5
        let concentrationValue = solutions[0].realConcentration!
        let molecularWeight = 90.0
        let volumesUsed = [2.40, 2.43,2.54]
        let expectedResults = [0.254626, 0.2578, 0.2695]
        
        let serie = session.series.first!
        
        let variables = session.experimentType.variables
        for variable in variables {
            var input = Input()
            switch variable.label{
            case Constants.sampleAcidMolecularWeight:
                input = Input(fromVariable: variable, withValue: molecularWeight)
                serie.addInput(input)
            case Constants.sampleWeight:
                input = Input(fromVariable: variable, withValue: weightInGrams)
                serie.addInput(input)
            case Constants.valorantConcentration:
                input = Input(fromVariable: variable, withValue: concentrationValue.value)
                serie.addInput(input)
            case Constants.valorantVolume:
                for volume in volumesUsed {
                    input = Input(fromVariable: variable, withValue: volume)
                    serie.addInput(input)
                }
                
                
            default:
                break
            }
            
            
        }
        // here we have introduced the values for the experiment and can request results
        let results = serie.getResultValuesForExperiment(experiment: session.experimentType)
        // Customary formatter for results
        let formatter = NumberFormatter()
        formatter.minimumIntegerDigits = 1
        formatter.maximumFractionDigits = 4
        
        let resultsPattern = results.map({(formatter.string(from: NSNumber(value:$0)))!})
//        let resultPattern = (formatter.string(from: NSNumber(value:results[0])))!
//        let expectedResultPattern = (formatter.string(from: NSNumber(value:expectedResult[0])))!
        let expectedResultsPatterns = expectedResults.map({(formatter.string(from: NSNumber(value:$0)))!})
        
        for (index,result) in resultsPattern.enumerated() {
            XCTAssertEqual(result, expectedResultsPatterns[index], "Incorrect results obtained->\(result) instead of \(expectedResultsPatterns[index])")

        }
        
    }
    
    func testStatisticsMultipleInputsForIogurt() {
        // This is an example of a functional test case.
        // Use XCTAssert and related functions to verify your tests produce the correct results.
        let weightInGrams = 8.5
        let concentrationValue = solutions[0].realConcentration!
        let molecularWeight = 90.0
        let volumesUsed = [2.40, 2.43,2.54]
        let expectedMean = 0.2606
        let expectedStdDev = 0.0078
        let expectedSEM = 0.0045
        
        let serie = session.series.first!
        
        let variables = session.experimentType.variables
        for variable in variables {
            var input = Input()
            switch variable.label{
            case Constants.sampleAcidMolecularWeight:
                input = Input(fromVariable: variable, withValue: molecularWeight)
                serie.addInput(input)
            case Constants.sampleWeight:
                input = Input(fromVariable: variable, withValue: weightInGrams)
                serie.addInput(input)
            case Constants.valorantConcentration:
                input = Input(fromVariable: variable, withValue: concentrationValue.value)
                serie.addInput(input)
            case Constants.valorantVolume:
                for volume in volumesUsed {
                    input = Input(fromVariable: variable, withValue: volume)
                    serie.addInput(input)
                }
                
                
            default:
                break
            }
            
            
        }
        // here we have introduced the values for the experiment and can request results
        let _ = serie.getResultValuesForExperiment(experiment: session.experimentType)
        let calculatedMean = serie.getMean() ?? 0.0
        let calculatedStdDev = serie.getStdDev() ?? 0.0
        let calculatedSEM = serie.getSEM() ?? 0.0
        print("Obtained values :\nMean= \(calculatedMean)\nStdDev= \(calculatedStdDev)")
        
        
        // Customary formatter for results
        let formatter = NumberFormatter()
        formatter.minimumIntegerDigits = 1
        formatter.maximumFractionDigits = 4
        
        // calculated values
        let calculatedMeanPattern = (formatter.string(from: NSNumber(value:calculatedMean)))!
        let calculatedStdDevPattern = (formatter.string(from: NSNumber(value:calculatedStdDev)))!
        let calculatedSEMPattern = (formatter.string(from: NSNumber(value:calculatedSEM)))!
        
        //Expected values
        let expectedMeanPattern = (formatter.string(from: NSNumber(value:expectedMean)))!
        let expectedStdDevPattern = (formatter.string(from: NSNumber(value:expectedStdDev)))!
        let expectedSEMPattern = (formatter.string(from: NSNumber(value:expectedSEM)))!
        
        XCTAssertEqual(calculatedMeanPattern, expectedMeanPattern, "Incorrect Mean calculation -> \(calculatedMeanPattern) instead of \(expectedMeanPattern)")
        XCTAssertEqual(calculatedStdDevPattern, expectedStdDevPattern, "Incorrect StdDev calculation -> \(calculatedStdDevPattern) instead of \(expectedStdDevPattern)")
        XCTAssertEqual(calculatedSEMPattern, expectedSEMPattern, "Incorrect Std error of the mean calculation -> \(calculatedSEMPattern) instead of \(expectedSEMPattern)")

        
    }
    
}
