//
//  MH_PerformanceTests.swift
//  
//
//  Created by Carlos Barrera on 6/12/17.
//

import XCTest
@testable import MeasurementHelper

class MH_PerformanceTests: XCTestCase {
    
    var session = Session()
    var solutions = [Solution]()

    enum Constants {
        static let valorantVolume = "titrantVolume"
        static let valorantConcentration = "titrantConcentration"
        static let sampleAcidMolecularWeight = "sampleAcidMolecularWeight"
        static let sampleWeight = "sampleWeight"
    }

    override func setUp() {
        super.setUp()
        // Put setup code here. This method is called before the invocation of each test method in the class.
        let solution1 = Solution(substance: "NaOH", atMolarity: 0.1, factor: 1.002, equivalence: 1, MW:nil, kind: .base)
        let solution2 = Solution(substance: "HCl", atMolarity: 0.01, factor: 0.998, equivalence: 1, MW:nil, kind: .acid)
        let solution3 = Solution(substance: "H2SO4", atMolarity: 0.01, factor: 1.001, equivalence: 2, MW:nil, kind: .acid)
        let solution4 = Solution(substance: "HNO3", atMolarity: 0.01, factor: 0.989, equivalence: 1, MW:nil, kind: .acid)
        let solution5 = Solution(substance: "KOH", atMolarity: 0.01, factor: 0.989, equivalence: 1, MW:nil, kind: .base)
        // create solutions collection
        solutions = [solution1, solution2, solution3, solution4, solution5]
        
        // Session
        // set startDate to now
        session.startDate = Date.init(timeIntervalSinceNow: 0.0)
        
        // Create experiment type
        let exp1 = Experiment()
        exp1.name = "Iogurt acidity"
        exp1.description = "Acid-Base titration to determine acidity of iogurt with NaOH"
        exp1.formula = Experiment.stepsFromFormula(formula: "\(Constants.valorantVolume) * \(Constants.valorantConcentration) * \(Constants.sampleAcidMolecularWeight) / \(Constants.sampleWeight) / 10 =")
        let var1 = Variable(label: Constants.valorantVolume, dim: .volume, shared: false, kind: .base)
        let var2 = Variable(label: Constants.valorantConcentration, dim: .concentration, shared: true, kind: .other)
        let var3 = Variable(label:Constants.sampleAcidMolecularWeight, dim: .concentration, shared: true, kind: .other)
        let var4 = Variable(label: Constants.sampleWeight, dim: .mass, shared: true, kind: .other)
        exp1.variables = [var1, var2, var3, var4]
        
        let result = Variable(label: "%AceticAcid", dim: .percentage, shared: false, kind: .other, description: "% of Acetic Acid in the sample", paired: false)
        exp1.result = result
        // set it as current experiment type
        session.experimentType = exp1
        let serie = Serie()
        
        serie.reactants.updateValue(solution1, forKey: solution1.substance)
        
        session.series.append(serie)
    }
    
    override func tearDown() {
        // Put teardown code here. This method is called after the invocation of each test method in the class.
        super.tearDown()
    }
    
    // function to generate random values arount a base value for testing
    func obtainRandomValuesAround(baseValue:Double, ammount:Int)-> [Double]{
        var result = [Double]()
        for _ in 0..<ammount {
            result.append(baseValue + Double((Int(arc4random_uniform(101))))/100.0)
        }
        return result
    }
    
    
    func testAddInputPerformance() {
        let weightInGrams = 8.5
        let concentrationValue = solutions[0].realConcentration!
        let molecularWeight = 90.0
        let numberOfInputs = 20
        let baseValue = 2.0
        let volumesUsed = obtainRandomValuesAround(baseValue: baseValue, ammount: numberOfInputs)
        
        let serie = session.series.first!
        
        let variables = session.experimentType.variables

        // let's see how long it takes to input those values, including the variables
        
        self.measure {
            // Put the code you want to measure the time of here.
            for variable in variables {
                var input = Input()
                switch variable.label{
                case Constants.sampleAcidMolecularWeight:
                    input = Input(fromVariable: variable, withValue: molecularWeight)
                    serie.addInput(input)
                case Constants.sampleWeight:
                    input = Input(fromVariable: variable, withValue: weightInGrams)
                    serie.addInput(input)
                case Constants.valorantConcentration:
                    input = Input(fromVariable: variable, withValue: concentrationValue.value)
                    serie.addInput(input)
                case Constants.valorantVolume:
                    for volume in volumesUsed {
                        input = Input(fromVariable: variable, withValue: volume)
                        serie.addInput(input)
                    }
                    
                    
                default:
                    break
                }
                
                
            }
        }
        // End of measure block
        
    }
    
    func testResultsCalculationPerformance() {
        let weightInGrams = 8.5
        let concentrationValue = solutions[0].realConcentration!
        let molecularWeight = 90.0
        let numberOfInputs = 20
        let baseValue = 2.0
        let volumesUsed = obtainRandomValuesAround(baseValue: baseValue, ammount: numberOfInputs)
        
        let serie = session.series.first!
        
        let variables = session.experimentType.variables
        for variable in variables {
            var input = Input()
            switch variable.label{
            case Constants.sampleAcidMolecularWeight:
                input = Input(fromVariable: variable, withValue: molecularWeight)
                serie.addInput(input)
            case Constants.sampleWeight:
                input = Input(fromVariable: variable, withValue: weightInGrams)
                serie.addInput(input)
            case Constants.valorantConcentration:
                input = Input(fromVariable: variable, withValue: concentrationValue.value)
                serie.addInput(input)
            case Constants.valorantVolume:
                for volume in volumesUsed {
                    input = Input(fromVariable: variable, withValue: volume)
                    serie.addInput(input)
                }
                
                
            default:
                break
            }
        }
        
        self.measure {
            let _ = serie.getResultValuesForExperiment(experiment: session.experimentType)
        }
    }
    
}
